#!/usr/bin/python3
# coding: utf-8

#################################################################
#
# Copyright (C) ИП Пуляев Григорий Васильевич, 2024
# email: rodegast@xmail.ru
# telegram: t.me/kybyc_meta_test
#
# Расширение добавляет поддержку браузера "Google Chrome"
# в платфому автоматизированного тестирования Meta Test.
#
# Этот файл НЕ является частью платфомы Meta Test.
# Лицензия расширения - Apache License 2.0
# https://www.apache.org/licenses/LICENSE-2.0
#
#################################################################

import os
import requests

from lxml import etree
from functools import lru_cache

from selenium import webdriver
from selenium.webdriver.chrome.service import Service as ChromeService

from core_brower import WebBrowser, install
from errors import InstallError, InstallExistsError
from parameter import CONFIG, LOG_TYPES, LOG_LEVEL, ItemConfig
from utilites import Version, get_platform, get_api_data, get_api_json

__all__ = (
	"get_chromedriver",
)

@lru_cache
def get_latest_version_chromedriver():
	"""
	Возвращает последнюю стабильную версию chromedriver-а
	"""
	try:
		data = get_api_json("https://googlechromelabs.github.io/chrome-for-testing/last-known-good-versions-with-downloads.json", timeout=CONFIG.REMOTE_TIMEOUT)
	except Exception as err:
		raise InstallError(err)
	if data:
		try:
			return data["channels"]["Stable"]["version"]
		except KeyError:
			raise InstallError("Невозможно получить последнюю версию chromedriver")
	raise InstallError("Невозможно получить последнюю версию chromedriver")

@lru_cache
def get_url_chromedriver(version, platform):
	"""
	Возввращает URL на скачивание драйвера
	"""
	short_version = ".".join(version.split(".")[:2])
	if Version(version) < Version(114):
		data = get_api_data("https://chromedriver.storage.googleapis.com/", timeout=CONFIG.REMOTE_TIMEOUT)
		if data:
			data = tuple( x[0].text for x in etree.fromstring(data) if "Contents" in x.tag )
			data = sorted([ x for x in data if platform in x and x.startswith(short_version) ], key=lambda x: len(x.split("/")[0]))
			if data:
				return "https://chromedriver.storage.googleapis.com/"+data[-1]
	else:
		data = get_api_json("https://googlechromelabs.github.io/chrome-for-testing/known-good-versions-with-downloads.json", timeout=CONFIG.REMOTE_TIMEOUT)
		for x in ( x["downloads"]["chromedriver"] for x in data["versions"] if ".".join(x["version"].split(".")[:2]) == short_version ):
			for y in x:
				if y["platform"] == platform:
					return y["url"]
	raise InstallError("Невозможно получить chromedriver для версии "+version)

@lru_cache
def get_chromedriver(version, browser_name):
	"""
	Возвращает путь к файлу chromedriver-а
	"""
	if not CONFIG.CHROME_DRIVER_PATH:
		if version == "latest":
			version = get_latest_version_chromedriver()
		platform      = get_platform()
		install_path  = os.path.join(CONFIG.BROWSERS_DIR, browser_name, version)
		driver_file   = os.path.join(install_path, "chromedriver")
		if platform == "win64":
			driver_file += ".exe"
		if not os.path.isfile(driver_file):
			url = get_url_chromedriver(version, platform)
			if platform == "win64":
				try:
					install(url, install_path, {"chromedriver-win64/chromedriver.exe": driver_file})
				except InstallExistsError as err:
					install(err.sourse, install_path, {"chromedriver.exe": driver_file})
			elif platform == "linux64":
				try:
					install(url, install_path, {"chromedriver-linux64/chromedriver": driver_file})
				except InstallExistsError as err:
					install(err.sourse, install_path, {"chromedriver": driver_file})
				os.chmod(driver_file, 0o4555)
			else:
				try:
					install(url, install_path, {"chromedriver-mac-x64/chromedriver": driver_file})
				except InstallExistsError as err:
					install(err.sourse, install_path, {"chromedriver": driver_file})
				os.chmod(driver_file, 0o4555)
		return driver_file
	return CONFIG.CHROME_DRIVER_PATH

class Chrome(WebBrowser):
	@lru_cache
	def installed_versions(self):
		"""
		Возвращает описание локально установленных браузеров
		{"версия браузера": "путь к исполняемому файлу"}
		"""
		platform = get_platform()
		if platform == "linux64":
			return self.linux_installed_version(("google-chrome", "google-chrome-stable", "google-chrome-beta", "google-chrome-dev"))
		elif platform == "mac-x64":
			return self.mac_installed_version("/Applications/Google Chrome.app/Contents/MacOS/Google Chrome")
		return self.windows_installed_version((
			r"PROGRAMFILES\\Google\\Chrome\\Application\\chrome.exe"
			, r"PROGRAMFILES(X86)\\Google\\Chrome\\Application\\chrome.exe"
			, r"LOCALAPPDATA\\Google\\Chrome\\Application\\chrome.exe"
			, r"PROGRAMW6432\\Google\\Chrome\\Application\\chrome.exe"
		))
	
	def linux_install_browser(self, install_path):
		"""
		Скачиваем и устанавливает браузер для linux
		Возвращает путь к исполняемому файлу
		"""
		chrome_path = os.path.join(install_path, "chrome-linux64", "chrome")
		if not os.path.isfile(chrome_path):
			url = "https://storage.googleapis.com/chrome-for-testing-public/%s/linux64/chrome-linux64.zip"%self.version
			install(url, install_path)
			os.chmod(chrome_path, 0o4555)
			crashpad_path = os.path.join(install_path, "chrome-linux64", "chrome_crashpad_handler")
			if os.path.isfile(crashpad_path):
				os.chmod(crashpad_path, 0o4555)
			sandbox_path = os.path.join(install_path, "chrome-linux64", "chrome_sandbox")
			if os.path.isfile(sandbox_path):
				os.chmod(sandbox_path, 0o4555)
		return chrome_path
	
	def windows_install_browser(self, install_path):
		"""
		Скачиваем и устанавливает браузер для windows
		Возвращает путь к исполняемому файлу
		"""
		chrome_path = os.path.join(install_path, "chrome-win64", "chrome.exe")
		if not os.path.isfile(chrome_path):
			url = "https://storage.googleapis.com/chrome-for-testing-public/%s/win64/chrome-win64.zip"%self.version
			install(url, install_path)
		return chrome_path
	
	def mac_install_browser(self, install_path):
		"""
		Скачиваем и устанавливает браузер для mac-64
		Возвращает путь к исполняемому файлу
		"""
		chrome_path = os.path.join(install_path, "chrome-mac-x64", "Google Chrome for Testing.app", "Contents", "MacOS", "Google Chrome for Testing")
		if not os.path.isfile(chrome_path):
			url = "https://storage.googleapis.com/chrome-for-testing-public/%s/mac-x64/chrome-mac-x64.zip"%self.version
			install(url, install_path)
			os.chmod(chrome_path, 0o4555)
			app_path = os.path.join(install_path, "chrome-mac-x64", "Google Chrome for Testing.app"
				, "Contents", "Frameworks", "Google Chrome for Testing Framework.framework"
				, "Versions", self.version
			)
			google_path = os.path.join(app_path, "Google Chrome for Testing Framework")
			if os.path.isfile(google_path):
				os.chmod(google_path, 0o4555)
			egl_path = os.path.join(app_path, "Libraries", "libEGL.dylib")
			if os.path.isfile(egl_path):
				os.chmod(egl_path, 0o4555)
			gles_path = os.path.join(app_path, "Libraries", "libGLESv2.dylib")
			if os.path.isfile(gles_path):
				os.chmod(gles_path, 0o4555)
			vk_path = os.path.join(app_path, "Libraries", "libvk_swiftshader.dylib")
			if os.path.isfile(vk_path):
				os.chmod(vk_path, 0o4555)
			app_path = os.path.join(app_path, "Helpers")
			crashpad_path = os.path.join(app_path, "chrome_crashpad_handler")
			if os.path.isfile(crashpad_path):
				os.chmod(crashpad_path, 0o4555)
			mode_path = os.path.join(app_path, "app_mode_loader")
			if os.path.isfile(mode_path):
				os.chmod(mode_path, 0o4555)
			shortcut_path = os.path.join(app_path, "web_app_shortcut_copier")
			if os.path.isfile(shortcut_path):
				os.chmod(shortcut_path, 0o4555)
			alert_path = os.path.join(app_path, "Google Chrome for Testing Helper (Alerts).app"
				, "Contents", "MacOS", "Google Chrome for Testing Helper (Alerts)"
			)
			if os.path.isfile(alert_path):
				os.chmod(alert_path, 0o4555)
			gpu_path = os.path.join(app_path, "Google Chrome for Testing Helper (GPU).app"
				, "Contents", "MacOS", "Google Chrome for Testing Helper (GPU)"
			)
			if os.path.isfile(gpu_path):
				os.chmod(gpu_path, 0o4555)
			plugin_path = os.path.join(app_path, "Google Chrome for Testing Helper (Plugin).app"
				, "Contents", "MacOS", "Google Chrome for Testing Helper (Plugin)"
			)
			if os.path.isfile(plugin_path):
				os.chmod(plugin_path, 0o4555)
			render_path = os.path.join(app_path, "Google Chrome for Testing Helper (Renderer).app"
				, "Contents", "MacOS", "Google Chrome for Testing Helper (Renderer)"
			)
			if os.path.isfile(render_path):
				os.chmod(render_path, 0o4555)
			helper_path = os.path.join(app_path, "Google Chrome for Testing Helper.app"
				, "Contents", "MacOS", "Google Chrome for Testing Helper"
			)
			if os.path.isfile(helper_path):
				os.chmod(helper_path, 0o4555)
		return chrome_path
	
	@lru_cache
	def available_versions(self):
		"""
		Возвращает список версий браузеров которые возможно установить
		"""
		try:
			data = get_api_json("https://googlechromelabs.github.io/chrome-for-testing/known-good-versions.json", timeout=CONFIG.REMOTE_TIMEOUT)
		except Exception:
			return []
		if data:
			return [ x["version"] for x in data.get("versions", []) ]
		return []
	
	def make_options(self):
		"""
		Возвращает объект конфигурации драйвера
		"""
		# https://peter.sh/experiments/chromium-command-line-switches/
		# https://gist.github.com/dodying/34ea4760a699b47825a766051f47d43b
		options = webdriver.ChromeOptions()
		options.add_experimental_option("excludeSwitches", ["enable-logging", "enable-automation"])
		options.add_experimental_option("useAutomationExtension", False)
		
		options.page_load_strategy = CONFIG.PAGE_LOAD_STRATEGY.lower()
		options.add_argument("--user-agent='Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/%s.0.0 Safari/537.36'"%self.short_version)
		options.add_argument("--silent")
		options.add_argument("--no-pings")
		options.add_argument("--no-sandbox")
		options.add_argument("--mute-audio")
		options.add_argument("--disable-fre")
		options.add_argument("--no-first-run")
		options.add_argument("--no-crash-upload")
		options.add_argument("--no-service-autorun")
		options.add_argument("--password-store=basic")
		options.add_argument("--disable-prompt-on-repost")
		options.add_argument("--disable-password-generation")
		options.add_argument("--disable-background-networking")
		
		options.add_argument("--ignore-ssl-errors=yes")
		options.add_argument("--ignore-certificate-errors")
		options.add_argument("--ignore-urlfetcher-cert-requests")
		
		options.add_argument("--disable-web-security")
		options.add_argument("--deny-permission-prompts")
		options.add_argument("--no-default-browser-check")
		options.add_argument("--allow-insecure-localhost")
		options.add_argument("--reduce-security-for-testing")
		options.add_argument("--disable-site-isolation-trials")
		options.add_argument("--allow-running-insecure-content")
		
		options.add_argument("--disable-breakpad")
		options.add_argument("--disable-infobars")
		options.add_argument("--disable-translate")
		options.add_argument("--disable-top-sites")
		options.add_argument("--disable-extensions")
		options.add_argument("--disable-auto-reload")
		options.add_argument("--disable-notifications")
		options.add_argument("--disable-dev-shm-usage")
		options.add_argument("--disable-popup-blocking")
		options.add_argument("--disable-crash-reporter")
		options.add_argument("--disable-plugins-discovery")
		options.add_argument("--disable-oopr-debug-crash-dump")
		options.add_argument("--disable-session-crashed-bubble")
		options.add_argument("--disable-ipc-flooding-protection")
		options.add_argument("--disable-blink-features=AutomationControlled")
		options.add_argument("--unsafely-allow-protected-media-identifier-for-domain")
		if Version(self.version) > Version(127):
			options.add_argument("--disable-search-engine-choice-screen")
		
		options.add_argument("--disable-features=IsolateOrigins,site-per-process,Translate"
			",InsecureDownloadWarnings,DownloadBubble,DownloadBubbleV2"
			",OptimizationTargetPrediction,OptimizationGuideModelDownloading"
			",SidePanelPinning,UserAgentClientHint,PrivacySandboxSettings4"
			",OptimizationHintsFetching,OptimizationHints,InterestFeedContentSuggestions"
		)
		
		if Version(self.version) > Version(117):
			options.add_argument("--remote-debugging-pipe")
		if self.enable_logs:
			options.set_capability("goog:loggingPrefs" if Version(self.version) > Version(75) else "loggingPrefs", {
				LOG_TYPES.BROWSER.lower()      : CONFIG.LOG_LEVEL
				, LOG_TYPES.CLIENT.lower()     : CONFIG.LOG_LEVEL
				, LOG_TYPES.DRIVER.lower()     : CONFIG.LOG_LEVEL
				, LOG_TYPES.SERVER.lower()     : CONFIG.LOG_LEVEL
				, LOG_TYPES.PROFILER.lower()   : CONFIG.LOG_LEVEL
				, LOG_TYPES.PERFORMANCE.lower(): CONFIG.LOG_LEVEL
			})
		if self.download_dir:
			# --unsafely-treat-insecure-origin-as-secure=http://yourhost:yourport
			prefs = {
				"profile.default_content_settings.popups": 0
				, "download.prompt_for_download": False
				, "download.directory_upgrade": True
				, "disable-popup-blocking": True
				, "safebrowsing.enabled": False
				, "plugins.always_open_pdf_externally": True
				, "plugins.plugins_disabled": ["Chrome PDF Viewer"]
				, "profile.default_content_setting_values.automatic_downloads": 1
			}
			options.enable_downloads = True
			if not self.is_local_browser:
				options.set_capability("se:downloadsEnabled", True)
			else:
				prefs["download.default_directory"] = self.download_dir
			options.add_experimental_option("prefs", prefs)
		else:
			options.enable_downloads = False
			options.set_capability("se:downloadsEnabled", False)
			if CONFIG.INCOGNITO:
				options.add_argument("--incognito")
		if self.cache_dir and not CONFIG.INCOGNITO:
			options.add_argument("--disk-cache-dir="+self.cache_dir)
		else:
			options.add_argument("--disable-cache")
			options.add_argument("--disk-cache-size=0")
			options.add_argument("--media-cache-size=0")
			options.add_argument("--disable-application-cache")
		if CONFIG.PROXY:
			options.add_argument("--proxy-server="+CONFIG.PROXY)
		if self.profile_dir:
			options.add_argument("--user-data-dir="+self.profile_dir)
			options.add_argument("--allow-profiles-outside-user-dir")
			options.add_argument("--enable-profile-shortcut-manager")
			options.add_argument("--profile-directory=Default")
		if CONFIG.HEADLESS and Version(self.version) > Version(58):
			if Version(self.version) > Version(108):
				options.add_argument("--headless=new")
			elif Version(self.version) > Version(96):
				options.add_argument("--headless=chrome")
			else:
				options.add_argument("--headless")
			
			options.add_argument("--disable-gpu")
			options.add_argument("--hide-scrollbars")
			options.add_argument("--disable-low-res-tiling")
			options.add_argument("--wm-window-animations-disabled")
			options.add_argument("--disable-renderer-backgrounding")
			options.add_argument("--disable-background-timer-throttling")
			options.add_argument("--disable-backgrounding-occluded-windows")
			options.add_argument("--disable-client-side-phishing-detection")
		
		if self.is_local_browser:
			if self.binary_location:
				options.binary_location = self.binary_location
		else:
			options.set_capability("browserName", self.browser_name)
			options.set_capability("browserVersion", self.version)
			options.set_capability("acceptInsecureCerts", True)
		return options
	
	def make_service(self):
		"""
		Возвращает объект обслуживания
		"""
		return ChromeService(self.driver_path)
	
	@property
	def web_driver_class(self):
		"""
		Возвращает класс веб-драйвера
		"""
		return webdriver.Chrome
	
	@property
	def enable_logs(self):
		"""
		Возвращает True если поддерживается логирование
		"""
		return True
	
	@property
	def user_version(self):
		"""
		Версия браузера которую указал пользователь в качестве используемой
		"""
		return CONFIG.CHROME_VERSION
	
	@property
	def user_local_path(self):
		"""
		Путь к исполняемому файлу браузера указанный пользователем
		"""
		return CONFIG.CHROME_LOCAL_PATH
	
	@property
	def browser_name(self):
		"""
		Возвращает имя браузера
		"""
		return "chrome"
	
	@property
	def driver_path(self):
		"""
		Возвращает путь к файлу драйвера
		"""
		return get_chromedriver(self.version, self.browser_name)

ItemConfig.registrator(
	ItemConfig("CHROME_DRIVER_PATH",  str, "", "Путь к chromedriver")
	, ItemConfig("CHROME_VERSION",    str, "", "CHROME: Версия chrome")
	, ItemConfig("CHROME_LOCAL_PATH", str, "", "CHROME: Путь к исполняемому файлу chrome")
)
