#!/usr/bin/python3
# coding: utf-8

#################################################################
#
# Copyright (C) ИП Пуляев Григорий Васильевич, 2024
# email: rodegast@xmail.ru
# telegram: t.me/kybyc_meta_test
#
# Расширение добавляет поддержку браузера Mozilla Firefox
# в платфому автоматизированного тестирования Meta Test.
#
# Этот файл НЕ является частью платфомы Meta Test.
# Лицензия расширения - Apache License 2.0
# https://www.apache.org/licenses/LICENSE-2.0
#
#################################################################

import os
from functools import lru_cache

from selenium import webdriver
from selenium.webdriver.firefox.firefox_profile import FirefoxProfile
from selenium.webdriver.firefox.service import Service as FirefoxService

from errors import InstallError
from core_brower import WebBrowser, install
from parameter import CONFIG, LOG_LEVEL, ItemConfig
from utilites import Version, get_platform, get_api_json, rMkDir

__all__ = (
	"get_geckodriver",
)

def get_driver_version(version):
	# https://github.com/SeleniumHQ/selenium/blob/trunk/common/geckodriver/geckodriver-support.json
	version = Version(version)
	if version >= Version("128.0"):
		return "0.36.0"
	elif version >= Version("115.0"):
		return "0.35.0"
	elif version >= Version("102.0"):
		return "0.33.0"
	elif version >= Version("91.0"):
		return "0.31.0"
	elif version >= Version("78.0"):
		return "0.30.0"
	elif version >= Version("60.0"):
		return "0.29.1"
	elif version >= Version("57.0"):
		return "0.25.0"
	elif version >= Version("55.0"):
		return "0.20.1"
	elif version >= Version("53.0"):
		return "0.18.0"
	elif version >= Version("52.0"):
		return "0.17.0"
	raise InstallError("Для версии %s недоступен geckodriver"%version)

@lru_cache
def get_geckodriver(version, browser_name):
	"""
	Возвращает путь к файлу geckodriver-а
	"""
	if not CONFIG.GECKO_DRIVER_PATH:
		platform     = get_platform()
		install_path = os.path.join(CONFIG.BROWSERS_DIR, browser_name, version)
		driver_file  = os.path.join(install_path, "geckodriver")
		if platform == "win64":
			driver_file += ".exe"
		if not os.path.isfile(driver_file):
			rMkDir(os.path.join(CONFIG.BROWSERS_DIR, browser_name))
			driver_version = get_driver_version(version)
			if platform == "win64":
				url = "https://github.com/mozilla/geckodriver/releases/download/v%s/geckodriver-v%s-win64.zip"%(driver_version, driver_version)
				install(url, install_path, {"geckodriver.exe": driver_file})
			elif platform == "linux64":
				url = "https://github.com/mozilla/geckodriver/releases/download/v%s/geckodriver-v%s-linux64.tar.gz"%(driver_version, driver_version)
				install(url, install_path, {"geckodriver": driver_file})
				os.chmod(driver_file, 0o4555)
			else:
				url = "https://github.com/mozilla/geckodriver/releases/download/v%s/geckodriver-v%s-macos.tar.gz"%(driver_version, driver_version)
				install(url, install_path, {"geckodriver": driver_file})
				os.chmod(driver_file, 0o4555)
		return driver_file
	return CONFIG.GECKO_DRIVER_PATH

class Firefox(WebBrowser):
	@lru_cache
	def installed_versions(self):
		"""
		Возвращает описание локально установленных браузеров
		{"версия браузера": "путь к исполняемому файлу"}
		"""
		platform = get_platform()
		if platform == "mac-x64":
			return self.mac_installed_version("/Applications/Firefox.app/Contents/MacOS/firefox")
		elif platform == "linux64":
			return self.linux_installed_version(("firefox", "firefox-esr"))
		return self.windows_installed_version((
			r"PROGRAMFILES\\Mozilla Firefox\\firefox.exe"
			, r"PROGRAMFILES(X86)\\Mozilla Firefox\\firefox.exe"
		))
	
	def linux_install_browser(self, install_path):
		"""
		Скачиваем и устанавливает браузер для linux
		Возвращает путь к исполняемому файлу
		"""
		firefox_path = os.path.join(install_path, "firefox", "firefox")
		if not os.path.isfile(firefox_path):
			url = "https://archive.mozilla.org/pub/firefox/releases/%s/linux-x86_64/ru/firefox-%s.tar.bz2"%(self.version, self.version)
			install(url, install_path)
			os.chmod(firefox_path, 0o4555)
		return firefox_path
	
	@lru_cache
	def available_versions(self):
		"""
		Возвращает список версий браузеров которые возможно установить
		"""
		try:
			data = get_api_json("https://product-details.mozilla.org/1.0/firefox_history_major_releases.json", timeout=CONFIG.REMOTE_TIMEOUT)
		except Exception as err:
			return []
		return list(data)
	
	def make_options(self):
		"""
		Возвращает объект конфигурации драйвера
		"""
		options = webdriver.FirefoxOptions()
		if self.profile_dir:
			profile = FirefoxProfile(profile_directory=self.profile_dir)
		else:
			profile = FirefoxProfile()
		profile.set_preference(
			"general.useragent.override"
			, "Mozilla/5.0 (X11; Linux x86_64; rv:%s) Gecko/20100101 Firefox/%s"%(self.short_version, self.short_version)
		)
		options.profile = profile
		options.add_argument("--disable-updater")
		options.page_load_strategy = CONFIG.PAGE_LOAD_STRATEGY.lower()
		if self.download_dir:
			options.enable_downloads = True
			profile.accept_untrusted_certs = True
			profile.set_eneble_native_elements = False
			profile.assume_untrusted_cert_issuer = False
			
			options.set_preference("pdfjs.disabled", True)
			options.set_preference("plugin.scan.Acrobat", "99.0")
			options.set_preference("plugin.scan.plid.all", False)
			options.set_preference("browser.download.folderList", 2)
			options.set_preference("browser.download.panel.shown", False)
			options.set_preference("browser.download.useToolkitUI", False)
			options.set_preference("browser.helperApps.alwaysAsk.force", False)
			options.set_preference("browser.download.manager.useWindow", False)
			options.set_preference("browser.download.animateNotifications", False)
			options.set_preference("browser.download.manager.closeWhenDone", True)
			options.set_preference("browser.download.manager.alertOnEXEOpen", False)
			options.set_preference("browser.download.manager.showWhenStarting", False)
			options.set_preference("browser.download.manager.focusWhenStarting", False)
			options.set_preference("browser.download.manager.showAlertOnComplete", False)
			
			if not self.is_local_browser:
				options.set_capability("se:downloadsEnabled", True)
			else:
				options.set_preference("browser.download.useDownloadDir", True)
				options.set_preference("browser.download.dir", self.download_dir)
			
		if CONFIG.HEADLESS and Version(self.version) > Version(55):
			options.add_argument("--headless")
		if CONFIG.PROXY:
			host, port = CONFIG.PROXY.split(":")
			options.set_preference("network.proxy.type", 1)
			options.set_preference("network.proxy.http", host)
			options.set_preference("network.proxy.http_port", int(port))
			options.set_preference("network.proxy.socks", host)
			options.set_preference("network.proxy.socks_port", int(port))
			options.set_preference("network.proxy.socks_remote_dns", False)
			options.set_preference("network.proxy.ssl", host)
			options.set_preference("network.proxy.ssl_port", int(port))
		if CONFIG.INCOGNITO:
			options.add_argument("--private")
			options.set_preference("browser.private.browsing.autostart", True)
		if self.cache_dir and not CONFIG.INCOGNITO:
			profile.set_preference("browser.cache.disk.enable", True)
			profile.set_preference("browser.cache.disk.parent_directory", self.cache_dir)
		else:
			profile.set_preference("browser.cache.disk.enable", False)
			profile.set_preference("browser.cache.memory.enable", True)
		if self.is_local_browser and self.binary_location:
			if self.binary_location:
				options.binary_location = self.binary_location
				options.firefox_binary  = self.binary_location
			options.executable_path = self.driver_path
		else:
			options.set_capability("browserName", self.browser_name)
			options.set_capability("browserVersion", self.version)
			options.set_capability("acceptInsecureCerts", True)
		return options
	
	def make_service(self):
		"""
		Возвращает объект обслуживания
		"""
		return FirefoxService(self.driver_path)
	
	@property
	def web_driver_class(self):
		"""
		Возвращает класс веб-драйвера
		"""
		return webdriver.Firefox
	
	@property
	def browser_name(self):
		"""
		Возвращает имя браузера
		"""
		return "firefox"
	
	@property
	def driver_path(self):
		"""
		Возвращает путь к файлу драйвера
		"""
		return get_geckodriver(self.version, self.browser_name)
	
	@property
	def user_local_path(self):
		"""
		Путь к исполняемому файлу браузера указанный пользователем
		"""
		return CONFIG.FIREFOX_LOCAL_PATH
	
	@property
	def user_version(self):
		"""
		Версия браузера которую указал пользователь в качестве используемой
		"""
		return CONFIG.FIREFOX_VERSION

ItemConfig.registrator(
	ItemConfig("GECKO_DRIVER_PATH",    str, "", "Путь к geckodriver")
	, ItemConfig("FIREFOX_VERSION",    str, "", "FIREFOX: Версия firefox")
	, ItemConfig("FIREFOX_LOCAL_PATH", str, "", "FIREFOX: Путь к исполняемому файлу firefox")
)
